/*
  thorfinn thorfinn@binf.ku.dk 18dec 2012
  assumes zero index positions

  should remove mutexs, and make function thread safe. dragon daemon

  part of angsd using faidx
 */



#include "analysisFunction.h"
#include "shared.h"
#include "faidx.h"
#include "general.h"
#include "pthread.h"

pthread_mutex_t getFasta_mut1 = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t getFasta_mut2 = PTHREAD_MUTEX_INITIALIZER;

class getFasta : public general{
private:
  char *ancName;
  char *refName;
  char *getRef(int refId,int *posi,int lens,faidx_t *fai);
  char *getAnc(int refId,int *posi,int lens,faidx_t *fai);
  faidx_t *faiRef;
  faidx_t *faiAnc;
public:
  getFasta(argStruct *arguments);
  ~getFasta();
  void printArg(FILE *argFile);
  void getOptions(argStruct *arguments);
  void run(funkyPars  *pars);//plugs in reference and ancestral
  void print(funkyPars *pars);
  void clean(funkyPars *pars);

};


void getFasta::printArg(FILE *argFile){
  fprintf(argFile,"---------------\n%s:\n\n",__FILE__);
  fprintf(argFile,"\t-ref\t%s\t(afile.fasta)\n",refName);
  fprintf(argFile,"\t-anc\t%s\t(afile.fasta)\n",ancName);
  fprintf(argFile,"\tNB these fasta files should be indexed 'samtools faidx'\n");
  fprintf(argFile,"\n");

}

getFasta::~getFasta(){
  if(refName!=NULL)
    getRef(-1,NULL,0,faiRef);//<-last indicate a cleanup
  if(ancName!=NULL)
    getAnc(-1,NULL,0,faiAnc);//<-last indicate a cleanup

  if(refName!=NULL&&faiRef!=NULL)
    fai_destroy(faiRef);
  if(ancName!=NULL&&faiAnc!=NULL)
    fai_destroy(faiAnc);

}


void getFasta::getOptions(argStruct *arguments){

  ancName = angsd::getArg("-anc",ancName,arguments);
  refName = angsd::getArg("-ref",refName,arguments);

}

 getFasta::getFasta(argStruct *arguments){
  refName = NULL;
  ancName = NULL;

   if(arguments->argc==2){
     if(!strcmp(arguments->argv[1],"-ref")||!strcmp(arguments->argv[1],"-anc")){
       printArg(stdout);
       exit(0);
    }else
       return;
   }

   getOptions(arguments);
   printArg(arguments->argumentFile);
   if(refName!=NULL){
    faiRef = fai_load(refName);
    if(faiRef==NULL){
      fprintf(stderr,"problems reading the (-ref ) reference genome\n");
      exit(0);
    }
   }
   if(ancName!=NULL){
    faiAnc = fai_load(ancName);
    if(faiAnc==NULL){
      fprintf(stderr,"problems reading the (-ans ) ancestral genome\n");
      exit(0);
    }
   }
}

void getFasta::run(funkyPars *pars){
  if(pars->numSites==0)
    return;
  if(refName!=NULL)
    pars->ref = getRef(pars->refId,pars->posi,pars->numSites,faiRef);
  if(ancName!=NULL)
    pars->anc = getAnc(pars->refId,pars->posi,pars->numSites,faiAnc);
  
}


char *getFasta::getRef(int refId, int* posi,int lens,faidx_t *fai) {
  pthread_mutex_lock(&getFasta_mut1);
  static char *ref = NULL;
  static int ref_len = -1;
  static int missingChrInFai =0;
  static int curChr = -1;
  if(refId==-1){
    free(ref);
    pthread_mutex_unlock(&getFasta_mut1);
    return NULL;
  }

  //  fprintf(stderr,"\t-> Reading fasta ref/anc [%s] fname=%s\n",__FUNCTION__,fasta);

  char *ret=new char[lens];


  static int count=0;
  //read a fasta for a new chr
  if(curChr==-1||refId!=curChr){
    missingChrInFai =0;
    free(ref);
    curChr = refId;
    ref = faidx_fetch_seq(fai, header->name[curChr], 0, 0x7fffffff, &ref_len);
    if(ref==NULL){
      fprintf(stderr,"[%s] Error loading reference alleles 1\n",__FUNCTION__);
      missingChrInFai =1;
    }
  }
  for(int i=0;i<lens;i++){
    if(missingChrInFai==0){
      if(posi[i]>ref_len){
	fprintf(stderr,"[%s] Warning position is after fasta :%s pos=%d ref_len=%d\n",__FUNCTION__,header->name[curChr],posi[i],ref_len);
	if(posi[i]>ref_len+200){
	  fprintf(stderr,"[%s] Trying to access fasta efter end of chromsome+200:%s pos=%d ref_len=%d\n",__FUNCTION__,header->name[curChr],posi[i],ref_len);
	  fflush(stderr);
	  exit(0);
	}
	ret[i]=4;
	continue;
      }
    }
    if(missingChrInFai)
      ret[i] = 4;
    else
      ret[i] = refToInt[ref[posi[i]]];//offseting by one, to make in work like samtools

  }
  pthread_mutex_unlock(&getFasta_mut1);

  return ret;

}


char *getFasta::getAnc(int refId, int* posi,int lens,faidx_t *fai) {
  pthread_mutex_lock(&getFasta_mut2);

  static char *ref = NULL;
  static int ref_len = -1;
  static int missingChrInFai =0;
  static int curChr = -1;
  if(refId==-1){
    free(ref);
    pthread_mutex_unlock(&getFasta_mut2);
    return NULL;
  }

  //  fprintf(stderr,"\t-> Reading fasta ref/anc [%s] fname=%s\n",__FUNCTION__,fasta);

  char *ret=new char[lens];
  //read a fasta for a new chr
  if(curChr==-1||refId!=curChr){
    missingChrInFai =0;
    free(ref);
    curChr = refId;
    ref = faidx_fetch_seq(fai, header->name[curChr], 0, 0x7fffffff, &ref_len);
    if(ref==NULL){
      fprintf(stderr,"[%s] Error loading reference alleles 1\n",__FUNCTION__);
      missingChrInFai =1;
    }
  }
  for(int i=0;i<lens;i++){
    if(missingChrInFai==0){
      if(posi[i]>ref_len){
	fprintf(stderr,"[%s] Warning position is after fasta :%s pos=%d ref_len=%d\n",__FUNCTION__,header->name[curChr],posi[i],ref_len);
	if(posi[i]>ref_len+200){
	  fprintf(stderr,"[%s] Trying to access fasta efter end of chromsome+200:%s pos=%d ref_len=%d\n",__FUNCTION__,header->name[curChr],posi[i],ref_len);
	  fflush(stderr);
	  exit(0);
	}
	ret[i]=4;
	continue;
      }
    }
    if(missingChrInFai)
      ret[i] = 4;
    else
      ret[i] = refToInt[ref[posi[i]]];//offseting by one, to make in work like samtools

  }
  pthread_mutex_unlock(&getFasta_mut2);

  return ret;

}

void getFasta::print(funkyPars *f){


}
void getFasta::clean(funkyPars *f){


}
